import json
from pathlib import Path
import pytest
from broker.broker import Broker
from broker.helpers import MockStub
from broker.providers.container import Container
from broker.settings import settings


class ContainerApiStub(MockStub):
    """This class stubs out the methods of the Container API client

    stubbing for:
     - self.runtime.images
     - self.runtime.pull_image(name)
     - self.runtime.image_info(image)["config"]["ExposedPorts"].items()
     - self.runtime.containers
     - self.runtime.get_attrs(cont_inst)
     - helpers.yaml_format(self.runtime.image_info(image))
     - self.runtime.create_container(container_host, **kwargs)
     - self.runtime.get_logs(cont_inst)
    """

    def __init__(self, **kwargs):
        in_dict = {
            "images": [MockStub({"tags": "ch-d:ubi8"})],  # self.runtime.images
            "containers": [MockStub({"tags": "f37d3058317f"})],  # self.runtime.containers
            "name": "f37d3058317f",  # self.runtime.get_attrs(cont_inst)["name"]
            "ports": MockStub({'22/tcp': [{'HostIp': '', 'HostPort': '1337'}]}),  # self.runtime.get_attrs(cont_inst)["ports"]
        }
        if "job_id" in kwargs:
            # we need to load in an image object
            super().__init__(self._load_image(kwargs.pop("job_id")))
        elif "name" in kwargs:
            # we need to load in a container object
            super().__init__(self._load_container(kwargs.pop("name")))
        else:
            super().__init__(in_dict=in_dict, **kwargs)

    @property
    def networks(self):
        return [MockStub(obj) for obj in json.loads(Path("tests/data/container/fake_networks.json").read_text())]

    def get_network_by_attrs(self, attr_dict):
        """Return the first matching network that matches all attr_dict keys and values."""
        for network in self.networks:
            if all(network.attrs.get(k) == v for k, v in attr_dict.items()):
                return network

    @staticmethod
    def pull_image(tag_name):
        with open("tests/data/container/fake_images.json") as image_file:
            image_data = json.load(image_file)
        for image in image_data:
            if tag_name in image["RepoTags"]:
                return MockStub(image)
        raise Broker.ProviderError(f"Unable to find image: {tag_name}")

    def create_container(self, container_host, **kwargs):
        if net_name := settings.container.network:
            net_dict = {}
            for name in net_name.split(","):
                if not self.get_network_by_attrs({"name": name}):
                    raise Exception(
                        f"Network '{name}' not found on container host."
                    )
                net_dict[name] = {"NetworkId": name}
            kwargs["networks"] = net_dict
        with open("tests/data/container/fake_containers.json") as container_file:
            container_data = json.load(container_file)
        image_data = self.pull_image(container_host)
        for container in container_data:
            if container["Config"]["Image"] == image_data.RepoTags[0]:
                container["id"] = container["Id"]  # hostname = cont_inst.id[:12]
                container["kwargs"] = kwargs
                return MockStub(container)


@pytest.fixture
def api_stub():
    return ContainerApiStub()


@pytest.fixture
def container_stub(api_stub):
    return Container(bind=api_stub)


def test_empty_init():
    assert Container()


def test_host_creation(container_stub):
    bx = Broker()
    cont = container_stub.run_container(container_host="ch-d:ubi8")
    host = container_stub.construct_host(cont, bx.host_classes)
    assert isinstance(host, bx.host_classes["host"])
    assert host.hostname == "f37d3058317f"


def test_single_network(container_stub):
    settings.container.network = "podman2"
    cont = container_stub.run_container(container_host="ch-d:ubi8")
    assert cont.kwargs["networks"] == {'podman2': {'NetworkId': 'podman2'}}


def test_multiple_networks(container_stub):
    settings.container.network = "podman1,podman2"
    cont = container_stub.run_container(container_host="ch-d:ubi8")
    assert cont.kwargs["networks"] == {'podman1': {'NetworkId': 'podman1'}, 'podman2': {'NetworkId': 'podman2'}}


def test_image_lookup_failure(container_stub):
    with pytest.raises(Broker.ProviderError) as err:
        container_stub.run_container(container_host="this-does-not-exist")
    assert "Unable to find image: this-does-not-exist" in err.value.message



"""Broker settings module.

Useful items:
    settings: The settings object.
    init_settings: Function to initialize the settings file.
    validate_settings: Function to validate the settings file.
    INTERACTIVE_MODE: Whether or not Broker is running in interactive mode.
    BROKER_DIRECTORY: The directory where Broker looks for its files.
    TEST_MODE: Whether or not Broker is running in a pytest session.
    settings_path: The path to the settings file.
    inventory_path: The path to the inventory file.
"""

import os
from pathlib import Path

import click
from dynaconf import Dynaconf, Validator
from dynaconf.validator import ValidationError

from broker.config_manager import ConfigManager
from broker.exceptions import ConfigurationError

INTERACTIVE_MODE = ConfigManager.interactive_mode
BROKER_DIRECTORY = Path.home().joinpath(".broker")
TEST_MODE = os.environ.get("BROKER_TEST_MODE", False)

if TEST_MODE:  # when in test mode, don't use the real broker directory
    BROKER_DIRECTORY = Path("tests/data/")
elif "BROKER_DIRECTORY" in os.environ:
    envar_location = Path(os.environ["BROKER_DIRECTORY"])
    if envar_location.is_dir():
        BROKER_DIRECTORY = envar_location

# ensure the broker directory exists
BROKER_DIRECTORY.mkdir(parents=True, exist_ok=True)

settings_path = BROKER_DIRECTORY.joinpath("broker_settings.yaml")
inventory_path = BROKER_DIRECTORY.joinpath("inventory.yaml")
cfg_manager = ConfigManager(settings_path)


if cfg_manager._get_migrations() and not TEST_MODE:
    if INTERACTIVE_MODE:
        click.secho(
            "Broker settings file has pending migrations.\n"
            "Continuing without running the migrations may cause errors.",
            fg="red",
        )
        if click.confirm("Would you like to run the migrations now?"):
            cfg_manager.migrate()
        else:
            click.secho("Continuing without running migrations.", fg="yellow")
    else:
        cfg_manager.migrate()

validators = [
    Validator("SSH", is_type_of=dict),
    Validator("SSH.HOST_USERNAME", default="root"),
    Validator("SSH.HOST_PASSWORD", default="toor"),
    Validator("SSH.HOST_CONNECTION_TIMEOUT", default=60),
    Validator("SSH.HOST_SSH_PORT", default=22),
    Validator("SSH.HOST_SSH_KEY_FILENAME", default=None),
    Validator("SSH.HOST_IPV6", default=False),
    Validator("SSH.HOST_IPV4_FALLBACK", default=True),
    Validator("SSH.BACKEND", default="ssh2-python312"),
    Validator("LOGGING", is_type_of=dict),
    Validator(
        "LOGGING.CONSOLE_LEVEL",
        is_in=["error", "warning", "info", "debug", "trace", "silent"],
        default="info",
    ),
    Validator(
        "LOGGING.FILE_LEVEL",
        is_in=["error", "warning", "info", "debug", "trace", "silent"],
        default="debug",
    ),
    Validator("THREAD_LIMIT", default=None),
    Validator("INVENTORY_FIELDS", is_type_of=dict),
    Validator("INVENTORY_LIST_VARS", is_type_of=str, default="hostname | name"),
    Validator("LESS_COLORS", default=False),
]

# temporary fix for dynaconf #751
vault_vars = {k: v for k, v in os.environ.items() if "VAULT_" in k}
for k in vault_vars:
    del os.environ[k]

settings = Dynaconf(
    settings_file=str(settings_path.absolute()),
    ENVVAR_PREFIX_FOR_DYNACONF="BROKER",
    validators=validators,
)
# to make doubly sure, remove the vault loader if set somehow
settings._loaders = [loader for loader in settings._loaders if "vault" not in loader]

try:
    settings.validators.validate(only="LOGGING")
except ValidationError as err:
    raise ConfigurationError(
        f"Configuration error in {settings_path.absolute()}: {err.args[0]}"
    ) from err

os.environ.update(vault_vars)



from contextlib import nullcontext
import json

import pytest

from broker import Broker
from broker.binds.foreman import ForemanBind
from broker.exceptions import AuthenticationError, ForemanBindError, ProviderError
from broker.helpers import MockStub
from broker.providers.foreman import Foreman

HOSTGROUP_VALID = "hg1"
HOSTGROUP_INVALID = "hg7"


class ForemanApiStub(MockStub, ForemanBind):
    """Runtime to mock queries to Foreman."""
    def __init__(self, **kwargs):
        MockStub.__init__(self, in_dict={})
        ForemanBind.__init__(self)

    def _post(self, url, **kwargs):
        if "/api/job_invocations" in url:
            command_slug = kwargs['json']['job_invocation']['inputs']['command'].split(" ")[0]
            with open("tests/data/foreman/fake_jobs.json") as jobs_file:
                job_data = json.load(jobs_file)
            return job_data[command_slug]
        if "/api/hosts" in url:
            with open("tests/data/foreman/fake_hosts.json") as hosts_file:
                host_data = json.load(hosts_file)
                return host_data
        print(url)

    def _get(self, url):
        with open("tests/data/foreman/fake_get.json") as file_:
            data = json.load(file_)
        try:
            return data[url]
        except:
            raise ProviderError(
                provider=self.__class__.__name__,
                message=f"Could not find endpoint {url}",
            )

    def _delete(self, url, **kwargs):
        print(url)

    def obtain_id_from_name(self, resource_type, resource_name):
        with open("tests/data/foreman/fake_resources.json") as resources_file:
            resources_data = json.load(resources_file)
        try:
            return resources_data[resource_type][resource_name]
        except:
            raise ProviderError(
                provider=self.__class__.__name__,
                message=f"Could not find {resource_name} in {resource_type}",
            )

    def job_output(self, job_id):
        jobs = {
                42: "success",
                43: "simple",
                44: "complex",
                45: "hostname",
                46: "complex-fail",
                }

        with open("tests/data/foreman/fake_jobs.json") as jobs_file:
            job_data = json.load(jobs_file)
        job_key = jobs[job_id]

        return job_data[job_key]["result"]

    def wait_for_job_to_finish(self, job_id):
        return

    def wait_for_host_to_install(self, hostname):
        return


@pytest.fixture
def api_stub():
    return ForemanApiStub()


@pytest.fixture
def foreman_stub(api_stub):
    return Foreman(bind=ForemanApiStub)


def test_empty_init():
    assert Foreman(bind=ForemanApiStub)


def test_inventory(foreman_stub):
    inventory = foreman_stub.get_inventory()
    assert len(inventory) == 2
    assert inventory[1]["name"] == "host2.fq.dn"
    assert inventory[0]["ip"] == "1.2.3.4"


def test_positive_host_creation(foreman_stub):
    new_host = foreman_stub.create_host(hostgroup=HOSTGROUP_VALID)
    assert new_host["name"] == "broker.local"
    assert new_host["mac"] == "00:11:22:33:44:55"


def test_negative_host_creation(foreman_stub):
    with pytest.raises(ProviderError):
        foreman_stub.create_host(hostgroup=HOSTGROUP_INVALID)


def test_positive_host(foreman_stub):
    bx = Broker()
    new_host = foreman_stub.create_host(hostgroup=HOSTGROUP_VALID)
    host = foreman_stub.construct_host(new_host, bx.host_classes)

    assert isinstance(host, bx.host_classes["host"])
    assert host.hostname == "broker.local"


def test_positive_remote_execution(foreman_stub):
    bx = Broker()
    new_host = foreman_stub.create_host(hostgroup=HOSTGROUP_VALID)
    host = foreman_stub.construct_host(new_host, bx.host_classes)

    result = host.execute("success")
    complex_result = host.execute("complex")

    assert result.status == 0
    assert complex_result.status == 0


def test_negative_remote_execution(foreman_stub):
    bx = Broker()
    new_host = foreman_stub.create_host(hostgroup=HOSTGROUP_VALID)
    host = foreman_stub.construct_host(new_host, bx.host_classes)

    simple_result = host.execute("simple")
    complex_result = host.execute("complex-fail")

    assert simple_result.status == 1
    assert complex_result.status == 100


@pytest.mark.parametrize(
    "response,expected,context",
    [
        ({1: 2}, {1: 2}, nullcontext()),
        (
            {"error": {"message": "Unable to authenticate user"}},
            None,
            pytest.raises(AuthenticationError),
        ),
        (
            {"error": {"full_messages": "foo", "message": "bar"}},
            None,
            pytest.raises(ForemanBindError),
        ),
        ({"errors": {"base": "foo"}}, None, pytest.raises(ForemanBindError)),
        ({"error": {"full_messages": ["bar"]}}, None, pytest.raises(ForemanBindError)),
    ],
)
def test_interpret_response(response, expected, context):
    bind = ForemanBind(
        foreman_username="",
        foreman_password="",
        url="",
        prefix="",
        verify=False,
    )

    with context:
        assert bind._interpret_response(response) == expected



"""Test file for broker.config_manager module."""
from ruamel.yaml import YAML
from broker.config_manager import ConfigManager, GH_CFG
from broker.settings import settings_path


yaml = YAML()
TEST_CFG_DATA = yaml.load(settings_path)


def test_basic_assertions():
    """Test ConfigManager class initialization and basic attributes."""
    cfg_mgr = ConfigManager(settings_path)
    assert isinstance(cfg_mgr._cfg, dict)
    assert cfg_mgr._cfg == TEST_CFG_DATA
    assert cfg_mgr.interactive_mode is False


def test_import_config():
    """ConfigManager should be able to download examples settings."""
    cfg_mgr = ConfigManager()
    result = cfg_mgr._import_config(GH_CFG, is_url=True)
    assert isinstance(result, str)
    converted = yaml.load(result)
    assert isinstance(converted, dict)

def test_get_e2e():
    """We should be able to get config chunks."""
    cfg_mgr = ConfigManager(settings_path)
    whole_cfg = cfg_mgr.get()
    assert isinstance(whole_cfg, dict)
    assert whole_cfg == TEST_CFG_DATA
    # get a speficic chunk
    logging_cfg = cfg_mgr.get('logging')
    assert logging_cfg == TEST_CFG_DATA['logging']
    # get a nested chunk
    test_nick = cfg_mgr.get('nicks.test_nick')
    assert test_nick == TEST_CFG_DATA['nicks']['test_nick']


def test_update_e2e():
    """We should be able to update config chunks."""
    cfg_mgr = ConfigManager(settings_path)
    # change logging level
    cfg_mgr.update('logging.console_level', 'debug')
    assert cfg_mgr.get('logging.console_level') == 'debug'
    # ensure a backup was created
    assert settings_path.with_suffix('.bak').exists()
    # restore original config and make sure the value is reverted
    cfg_mgr.restore()
    # load a new instance of ConfigManager to ensure the change was reverted
    cfg_mgr = ConfigManager(settings_path)
    assert cfg_mgr.get('logging.console_level') == TEST_CFG_DATA['logging']['console_level']


def test_nicks():
    """Specifically test the nick functionality."""
    cfg_mgr = ConfigManager(settings_path)
    nick_list = cfg_mgr.nicks()
    assert "test_nick" in nick_list
    test_nick = cfg_mgr.nicks("test_nick")
    assert test_nick == TEST_CFG_DATA['nicks']['test_nick']
